# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Combo box with a chevron selector
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2011-2024, Damon Lynch"

from PyQt5.QtCore import QPointF, QSize, Qt
from PyQt5.QtGui import QFont, QFontMetrics, QPainter
from PyQt5.QtWidgets import QComboBox, QLabel, QSizePolicy

from raphodo.ui.viewutils import darkModePixmap


class ChevronCombo(QComboBox):
    """
    Combo box with a chevron selector
    """

    def __init__(self, in_panel: bool = False, parent=None) -> None:
        """
        :param in_panel: if True, widget color set to background color,
         else set to window color
        """
        super().__init__(parent)

    def paintEvent(self, event):
        painter = QPainter(self)

        # Draw chevron (down arrow)
        width = int(QFontMetrics(QFont()).height() * (2 / 3))
        size = QSize(width, width)
        pixmap = darkModePixmap(path=":/icons/chevron-down.svg", size=size)
        x = self.rect().width() - width - 6
        y = self.rect().center().y() - width / 2
        p = QPointF(x, y)
        painter.drawPixmap(p, pixmap)

        # Draw text
        painter.setPen(self.palette().windowText().color())
        painter.drawText(
            self.rect(), Qt.AlignVCenter | Qt.AlignLeft, self.currentText()
        )

    def makeLabel(self, text: str) -> QLabel:
        """
        Render a label to attach to this widget
        """
        label = QLabel(text)
        label.setAlignment(Qt.AlignBottom)
        label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Maximum)
        return label



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Toggle Switch reminiscent of Android On/off switches:
https://www.google.com/design/spec/components/selection-controls.html

Visual style is rounded. However, by adjusting the style sheet it can be
made like a rounded square, close to how Gnome handles it, albeit
without the "ON"/"OFF text.

Inspiration:
http://stackoverflow.com/questions/14780517/toggle-switch-in-qt
http://thesmithfam.org/blog/2010/03/10/fancy-qslider-stylesheet/
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"


from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtGui import QColor, QFont, QFontMetrics, QPalette
from PyQt5.QtWidgets import QAbstractSlider, QApplication, QSlider


class QToggleSwitch(QSlider):
    """
    Toggle Switch reminiscent of Android On/off switches.

    Connect to signal valueChanged to react to user setting the switch.
    """

    def __init__(
        self, background: QColor | None = None, parent=None, size: int = 2
    ) -> None:
        """
        Toggle switch that can be dragged or clicked to change value

        :param background: background color
        :param parent: parent widget
        :param size: size of widget as multiplier, where base widget height is half
         that of font height
        """
        super().__init__(Qt.Horizontal, parent)

        self.base_height = QFontMetrics(QFont()).height() // 2 * size
        self.radius = self.base_height // 2

        width = self.base_height * 2
        self.widgetWidth = width
        self.handleWidth = width // 2
        self.sliderRange = width
        self.sliderMidPoint = width // 2
        self.setRange(0, self.sliderRange)

        self.setMaximumWidth(self.widgetWidth)
        self.setFixedHeight(self.base_height + 6)

        # Track if button was dragged in the control
        self.dragged = False

        self.setStyleSheet(self.stylesheet(background))

        self.actionTriggered.connect(self.onActionTriggered)
        self.sliderReleased.connect(self.onSliderRelease)

    def stylesheet(self, background: QColor | None) -> str:
        shading_intensity = 104
        windowColor = QPalette().color(QPalette().Window)  # type: QColor

        if background is None:
            backgroundName = windowColor.name()
        else:
            backgroundName = QColor(background).name()

        handleLightName = (QPalette().color(QPalette().Light)).name()  # type: str
        handleDarkName = (QPalette().color(QPalette().Dark)).name()  # type: str
        handleHoverLightName = (
            (QPalette().color(QPalette().Light)).lighter(shading_intensity).name()
        )
        handleHoverDarkName = (
            (QPalette().color(QPalette().Dark)).darker(shading_intensity).name()
        )

        insetDarkName = windowColor.darker(108).name()
        insetLightName = windowColor.darker(102).name()

        highlightColor = QPalette().color(QPalette().Highlight)  # type: QColor
        highlightLightName = highlightColor.lighter(110).name()
        highlightDarkName = highlightColor.darker(130).name()

        return f"""
            QSlider::groove:horizontal {{
                background-color: {backgroundName};
                height: {self.base_height}px;
            }}

            QSlider::sub-page:horizontal {{
            background: qlineargradient(x1: 0, y1: 0.2, x2: 1, y2: 1,
                stop: 0 {highlightDarkName}, stop: 1 {highlightLightName});
            border: 1px solid #777;
            border-top-left-radius: {self.radius}px;
            border-bottom-left-radius: {self.radius}px;
            }}

            QSlider::add-page:horizontal {{
            background: qlineargradient(x1: 0, y1: 0.2, x2: 1, y2: 1,
                stop: 0 {insetDarkName}, stop: 1 {insetLightName});
            border: 1px solid #777;
            border-top-right-radius: {self.radius}px;
            border-bottom-right-radius: {self.radius}px;
            }}

            QSlider::handle:horizontal {{
            background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                stop:0 {handleLightName}, stop:1 {handleDarkName});
            border: 1px solid #777;
            width: {self.handleWidth}px;
            border-radius: {self.radius}px;
            }}

            QSlider::handle:horizontal:hover {{
            background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                stop:0 {handleHoverLightName}, stop:1 {handleHoverDarkName});
            border: 1px solid #444;
            border-radius: {self.radius}px;
            }}

            QSlider::sub-page:horizontal:disabled {{
            background: #bbb;
            border-color: #999;
            }}

            QSlider::add-page:horizontal:disabled {{
            background: #eee;
            border-color: #999;
            }}

            QSlider::handle:horizontal:disabled {{
            background: #eee;
            border: 1px solid #aaa;
            border-radius: {self.radius}px;
            }}
        """

    @pyqtSlot(int)
    def onActionTriggered(self, action: int) -> None:
        if action == QAbstractSlider.SliderMove:
            self.dragged = True
        else:
            if action % 2:
                self.setValue(self.sliderRange)
            else:
                self.setValue(0)

    @pyqtSlot()
    def onSliderRelease(self) -> None:
        if self.dragged:
            if self.sliderPosition() >= self.sliderMidPoint:
                self.setValue(self.sliderRange)
            else:
                self.setValue(0)
            self.dragged = False
        else:
            # Account for user pressing the button itself
            if self.value() == self.sliderRange:
                self.setValue(0)
            else:
                self.setValue(self.sliderRange)

    def on(self) -> bool:
        return self.value() == self.sliderRange

    def setOn(self, on: bool = True) -> None:
        if on:
            self.setValue(self.sliderRange)
        else:
            self.setValue(0)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    b = QToggleSwitch(size=10)
    b.show()
    sys.exit(app.exec_())



#!/usr/bin/env python3

# Copyright (C) 2011-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2011-2024, Damon Lynch"

import datetime
import logging
import os
import re
from pathlib import Path
from typing import NamedTuple

from packaging.version import Version, parse
from PyQt5.QtCore import QSettings, Qt, QTime

import raphodo.__about__
import raphodo.constants as constants
from raphodo.constants import FileType, PresetPrefType
from raphodo.generatenameconfig import (
    DEFAULT_PHOTO_RENAME_PREFS,
    DEFAULT_SUBFOLDER_PREFS,
    DEFAULT_VIDEO_RENAME_PREFS,
    DEFAULT_VIDEO_SUBFOLDER_PREFS,
    DICT_IMAGE_RENAME_L0,
    DICT_SUBFOLDER_L0,
    DICT_VIDEO_RENAME_L0,
    DICT_VIDEO_SUBFOLDER_L0,
    DOWNLOAD_SEQ_NUMBER,
    JOB_CODE,
    LIST_SEQUENCE_L1,
    LOWERCASE,
    PHOTO_RENAME_MENU_DEFAULTS_CONV,
    SEPARATOR,
    SEQUENCE_LETTER,
    SESSION_SEQ_NUMBER,
    STORED_SEQ_NUMBER,
    VIDEO_RENAME_MENU_DEFAULTS_CONV,
    PrefError,
    PrefValueKeyComboError,
    check_pref_valid,
    upgrade_pre090a4_rename_pref,
)
from raphodo.metadata.fileformats import ALL_KNOWN_EXTENSIONS
from raphodo.storage.storage import (
    get_media_dir,
    platform_photos_directory,
    platform_photos_identifier,
    platform_videos_directory,
    platform_videos_identifier,
)
from raphodo.utilities import (
    available_cpu_count,
    default_thumbnail_process_count,
    make_internationalized_list,
)


class ScanPreferences:
    r"""
    Handle user preferences while scanning devices like memory cards,
    cameras or the filesystem.

    Sets data attribute valid to True if ignored paths are valid. An ignored
    path is always assumed to be valid unless regular expressions are used.
    If regular expressions are used, then it is valid only if a valid
    regular expression can be compiled from each line.

    >>> no_ignored_paths = ScanPreferences([])
    >>> no_ignored_paths.valid
    True

    >>> some_paths = ScanPreferences(['.Trash', '.thumbnails'])
    >>> some_paths.valid
    True

    >>> some_re_paths = ScanPreferences(['.Trash', '\.[tT]humbnails'], True)
    >>> some_re_paths.valid
    True

    >>> some_more_re_paths = ScanPreferences(['.Trash', '\.[tThumbnails'], True)
    >>> some_more_re_paths.valid
    False
    """

    def __init__(
        self, ignored_paths: list[str], use_regular_expressions: bool = False
    ) -> None:
        self.ignored_paths = ignored_paths
        self.use_regular_expressions = use_regular_expressions

        if ignored_paths and use_regular_expressions:
            self.valid = self._check_and_compile_re()
        else:
            self.re_pattern = None
            self.valid = True

    def scan_this_path(self, path: str) -> bool:
        """
        Returns true if the path should be included in the scan.
        Assumes path is a full path
        """

        # see method list_not_empty() in Preferences class to see
        # what an "empty" list is: ['']
        if not (self.ignored_paths and self.ignored_paths[0]):
            return True

        if not self.use_regular_expressions:
            return not path.endswith(tuple(self.ignored_paths))

        return not self.re_pattern.match(path)

    def _check_and_compile_re(self) -> bool:
        """
        Take the ignored paths and attempt to compile a regular expression
        out of them. Checks line by line.

        :return: True if there were no problems creating the regular
        expression pattern
        """

        assert self.use_regular_expressions

        error_encountered = False
        pattern = ""
        for path in self.ignored_paths:
            # check path for validity
            try:
                re.match(path, "")
                pattern += f".*{path}s$|"
            except re.error:
                logging.error(f"Ignoring malformed regular expression: {path}")
                error_encountered = True

        if pattern:
            pattern = pattern[:-1]

            try:
                self.re_pattern = re.compile(pattern)
            except re.error:
                logging.error(f"This regular expression is invalid: {pattern}")
                self.re_pattern = None
                error_encountered = True

        logging.debug(f"Ignored paths regular expression pattern: {pattern}")

        return not error_encountered


class DownloadsTodayTracker:
    """
    Handles tracking the number of successful downloads undertaken
    during any one day.

    When a day starts is flexible. See for more details:
    http://damonlynch.net/rapid/documentation/#renameoptions
    """

    def __init__(self, downloads_today: list[str], day_start: str) -> None:
        """

        :param downloads_today: list[str,str] containing date and the
         number of downloads today e.g. ['2015-08-15', '25']
        :param day_start: the time the day starts, e.g. "03:00"
         indicates the day starts at 3 a.m.
        """
        self.day_start = day_start
        self.downloads_today = downloads_today

    def get_or_reset_downloads_today(self) -> int:
        """
        Primary method to get the Downloads Today value, because it
        resets the value if no downloads have already occurred on the
        day of the download.
        :return: the number of successful downloads that have occurred
        today
        """
        v = self.get_downloads_today()
        if v <= 0:
            self.reset_downloads_today()
            # -1 was returned in the Gtk+ version of Rapid Photo Downloader -
            # why?
            v = 0
        return v

    def get_downloads_today(self) -> int:
        """
        :return the preference value for the number of successful
        downloads performed today. If value is less than zero,
        the date has changed since the value was last updated.
        """

        hour, minute = self.get_day_start()
        try:
            adjusted_today = datetime.datetime.strptime(
                f"{self.downloads_today[0]} {hour}:{minute}", "%Y-%m-%d %H:%M"
            )
        except Exception:
            logging.critical(
                "Failed to calculate date adjustment. Download today values "
                "appear to be corrupted: %s %s:%s",
                self.downloads_today[0],
                hour,
                minute,
            )
            adjusted_today = None

        now = datetime.datetime.today()

        if adjusted_today is None:
            return -1

        if now < adjusted_today:
            try:
                return int(self.downloads_today[1])
            except ValueError:
                logging.error("Invalid Downloads Today value. Resetting value to zero.")
                self.reset_downloads_today()
                return 0
        else:
            return -1

    def get_day_start(self) -> tuple[int, int]:
        """
        :return: hour and minute components as tuple of ints
        """
        try:
            t1, t2 = self.day_start.split(":")
            return int(t1), int(t2)
        except ValueError:
            logging.error(
                "'Start of day' preference value %s is corrupted. Resetting "
                "to midnight",
                self.day_start,
            )
            self.day_start = "0:0"
            return 0, 0

    def increment_downloads_today(self) -> bool:
        """
        :return: True if day changed
        """
        v = self.get_downloads_today()
        if v >= 0:
            self.set_downloads_today(self.downloads_today[0], v + 1)
            return False
        else:
            self.reset_downloads_today(1)
            return True

    def reset_downloads_today(self, value: int = 0) -> None:
        now = datetime.datetime.today()
        hour, minute = self.get_day_start()
        t = datetime.time(hour, minute)
        if now.time() < t:
            date = today()
        else:
            d = datetime.datetime.today() + datetime.timedelta(days=1)
            date = d.strftime("%Y-%m-%d")

        self.set_downloads_today(date, value)

    def set_downloads_today(self, date: str, value: int = 0) -> None:
        self.downloads_today = [date, str(value)]

    def set_day_start(self, hour: int, minute: int) -> None:
        self.day_start = f"{hour}:{minute}"

    def log_vals(self) -> None:
        logging.info(
            "Date %s Value %s Day start %s",
            self.downloads_today[0],
            self.downloads_today[1],
            self.day_start,
        )


def today():
    return datetime.date.today().strftime("%Y-%m-%d")


class WSLWindowsDrivePrefs(NamedTuple):
    drive_letter: str
    label: str
    auto_mount: bool
    auto_unmount: bool


is_devel_env = os.getenv("RPD_DEVEL_DEFAULTS") is not None


class Preferences:
    """
    Program preferences, being a mix of user facing and non-user facing prefs.
    """

    program_defaults = dict(program_version="")
    rename_defaults = dict(
        photo_download_folder=platform_photos_directory(),
        video_download_folder=platform_videos_directory(),
        photo_subfolder=DEFAULT_SUBFOLDER_PREFS,
        video_subfolder=DEFAULT_VIDEO_SUBFOLDER_PREFS,
        photo_rename=DEFAULT_PHOTO_RENAME_PREFS,
        video_rename=DEFAULT_VIDEO_RENAME_PREFS,
        # following two extension values introduced in 0.9.0a4:
        photo_extension=LOWERCASE,
        video_extension=LOWERCASE,
        day_start="03:00",
        downloads_today=[today(), "0"],
        stored_sequence_no=0,
        strip_characters=True,
        synchronize_raw_jpg=False,
        job_codes=[_("Wedding"), _("Birthday")],
        remember_job_code=True,
        ignore_mdatatime_for_mtp_dng=True,
    )

    # custom preset prefs are define below in code such as get_preset()
    timeline_defaults = dict(proximity_seconds=3600)

    display_defaults = dict(
        detailed_time_remaining=False,
        warn_downloading_all=True,
        warn_backup_problem=True,
        warn_broken_or_missing_libraries=True,
        warn_fs_metadata_error=True,
        warn_unhandled_files=True,
        ignore_unhandled_file_exts=["TMP", "DAT"],
        job_code_sort_key=0,
        job_code_sort_order=0,
        did_you_know_on_startup=not is_devel_env,
        did_you_know_index=0,
        # see constants.CompletedDownloads:
        completed_downloads=3,
        consolidate_identical=True,
        # see constants.TreatRawJpeg:
        treat_raw_jpeg=2,
        # see constants.MarkRawJpeg:
        mark_raw_jpeg=3,
        # introduced in 0.9.6b1:
        auto_scroll=True,
        # If you change the language setting update it in __init__.py too, where it is
        # read directly without using this class.
        language="",
        show_system_folders=False,  # Introduced in 0.9.27b2
        survey_countdown=10,  # Introduced in 0.9.29
        survey_taken=2022 if is_devel_env else 0,  # Year. Introduced in 0.9.29
        never_prompt_for_survey=False,  # Introduced in 0.9.29
    )
    device_defaults = dict(
        only_external_mounts=True,
        device_autodetection=True,
        this_computer_source=False,
        this_computer_path="",
        scan_specific_folders=True,
        # pre 0.9.3a1 value: device_without_dcim_autodetection=False, is now replaced by
        # scan_specific_folders
        folders_to_scan=["DCIM", "PRIVATE", "MP_ROOT"],
        ignored_paths=[".Trash", ".thumbnails", "THMBNL", "__MACOSX", "Screenshots"],
        use_re_ignored_paths=False,
        volume_whitelist=[""],
        volume_blacklist=[""],
        camera_blacklist=[""],
    )
    backup_defaults = dict(
        backup_files=False,
        backup_device_autodetection=True,
        photo_backup_identifier=platform_photos_identifier(),
        video_backup_identifier=platform_videos_identifier(),
        backup_photo_location=os.path.expanduser("~"),
        backup_video_location=os.path.expanduser("~"),
    )
    automation_defaults = dict(
        auto_download_at_startup=False,
        auto_download_upon_device_insertion=False,
        auto_unmount=False,
        auto_exit=False,
        auto_exit_force=False,
        move=False,
        verify_file=False,
        auto_mount=True,  # new in 0.9.29a1
    )
    performance_defaults = dict(
        generate_thumbnails=True,
        use_thumbnail_cache=True,
        save_fdo_thumbnails=True,
        max_cpu_cores=default_thumbnail_process_count(),
        keep_thumbnails_days=30,
    )
    error_defaults = dict(
        conflict_resolution=int(constants.ConflictResolution.skip),
        backup_duplicate_overwrite=False,
    )
    destinations = dict(photo_backup_destinations=[""], video_backup_destinations=[""])
    version_check = dict(
        check_for_new_versions=True,
        include_development_release=False,
        ignore_versions=[""],
    )
    restart_directives = dict(purge_thumbnails=False, optimize_thumbnail_db=False)
    metadata_defaults = dict(
        force_exiftool=False,
        force_exiftool_video=False,  # new in 0.9.35
        ignore_time_zone_changes=True,  # new in 0.9.29a1
        time_zone_offset_resolution=60,  # new in 0.9.29a1
    )
    # New in 0.9.27b2:
    wsl_defaults = dict(
        wsl_automount_removable_drives=True,
        wsl_automount_all_removable_drives=False,
    )

    def __init__(self) -> None:
        # To avoid infinite recursions arising from the use of __setattr__,
        # manually assign class values to the class dict
        self.__dict__["settings"] = QSettings(
            "Rapid Photo Downloader", "Rapid Photo Downloader"
        )
        self.__dict__["valid"] = True

        # These next two values must be kept in sync
        dicts = (
            self.program_defaults,
            self.rename_defaults,
            self.timeline_defaults,
            self.display_defaults,
            self.device_defaults,
            self.backup_defaults,
            self.automation_defaults,
            self.performance_defaults,
            self.error_defaults,
            self.destinations,
            self.version_check,
            self.restart_directives,
            self.metadata_defaults,
            self.wsl_defaults,
        )
        group_names = (
            "Program",
            "Rename",
            "Timeline",
            "Display",
            "Device",
            "Backup",
            "Automation",
            "Performance",
            "ErrorHandling",
            "Destinations",
            "VersionCheck",
            "RestartDirectives",
            "Metadata",
            "WindowsSubsystemLinux",
        )
        assert len(dicts) == len(group_names)

        # Create quick lookup table for types of each value, including the
        # special case of lists, which use the type of what they contain.
        # While we're at it also merge the dictionaries into one dictionary
        # of default values.
        self.__dict__["types"] = {}
        self.__dict__["defaults"] = {}
        for d in dicts:
            for key, value in d.items():
                t = type(value[0]) if isinstance(value, list) else type(value)
                self.types[key] = t
                self.defaults[key] = value
        # Create quick lookup table of the group each key is in
        self.__dict__["groups"] = {}
        for idx, d in enumerate(dicts):
            for key in d:
                self.groups[key] = group_names[idx]

    def __getitem__(self, key):
        group = self.groups.get(key, "General")
        self.settings.beginGroup(group)
        v = self.settings.value(key, self.defaults[key], self.types[key])
        self.settings.endGroup()
        return v

    def __getattr__(self, key):
        return self[key]

    def __setitem__(self, key, value):
        group = self.groups.get(key, "General")
        self.settings.beginGroup(group)
        self.settings.setValue(key, value)
        self.settings.endGroup()

    def __setattr__(self, key, value):
        self[key] = value

    def value_is_set(self, key, group: str | None = None) -> bool:
        if group is None:
            group = "General"

        group = self.groups.get(key, group)
        self.settings.beginGroup(group)
        v = self.settings.contains(key)
        self.settings.endGroup()
        return v

    def sync(self):
        self.settings.sync()

    def status(self) -> QSettings.Status:
        return self.settings.status()

    def restore(self, key: str) -> None:
        self[key] = self.defaults[key]

    def get_custom_presets(
        self, preset_type: PresetPrefType
    ) -> tuple[list[str], list[list[str]]]:
        """
        Returns the custom presets for the particular type.

        :param preset_type: one of photo subfolder, video subfolder, photo
         rename, or video rename
        :return: tuple of list of present names and list of pref lists. Each
         item in the first list corresponds with the item of the same index in the
         second list.
        """

        preset_pref_lists = []
        preset_names = []

        self.settings.beginGroup("Presets")

        preset = preset_type.name
        size = self.settings.beginReadArray(preset)
        for i in range(size):
            self.settings.setArrayIndex(i)
            preset_names.append(self.settings.value("name", type=str))
            preset_pref_lists.append(self.settings.value("pref_list", type=str))
        self.settings.endArray()

        self.settings.endGroup()

        return preset_names, preset_pref_lists

    def set_custom_presets(
        self,
        preset_type: PresetPrefType,
        preset_names: list[str],
        preset_pref_lists: list[list[str]],
    ) -> None:
        """
        Saves a list of custom presets in the user's preferences.

        If the list of preset names is empty, the preference value will be cleared.

        :param preset_type: one of photo subfolder, video subfolder, photo
         rename, or video rename
        :param preset_names: list of names for each pref list
        :param preset_pref_lists: the list of pref lists
        """

        self.settings.beginGroup("Presets")

        preset = preset_type.name

        # Clear all the existing presets with that name.
        # If we don't do this, when the array shrinks, old values can hang around,
        # even though the array size is set correctly.
        self.settings.remove(preset)

        self.settings.beginWriteArray(preset)
        for i in range(len(preset_names)):
            self.settings.setArrayIndex(i)
            self.settings.setValue("name", preset_names[i])
            self.settings.setValue("pref_list", preset_pref_lists[i])
        self.settings.endArray()

        self.settings.endGroup()

    def get_wsl_drives(self) -> list[WSLWindowsDrivePrefs]:
        drives = []
        self.settings.beginGroup("WindowsSubsystemLinux")
        setting = "drives"
        size = self.settings.beginReadArray(setting)
        for i in range(size):
            self.settings.setArrayIndex(i)
            drive = self.settings.value("drive")
            drives.append(
                WSLWindowsDrivePrefs(
                    drive_letter=drive[0],
                    label=drive[1],
                    auto_mount=drive[2] == "true",
                    auto_unmount=drive[3] == "true",
                )
            )
        self.settings.endArray()
        self.settings.endGroup()
        return drives

    def set_wsl_drives(self, drives: list[WSLWindowsDrivePrefs]):
        self.settings.beginGroup("WindowsSubsystemLinux")
        setting = "drives"
        self.settings.remove(setting)
        self.settings.beginWriteArray(setting)
        for i in range(len(drives)):
            self.settings.setArrayIndex(i)
            drive = drives[i]
            self.settings.setValue(
                "drive",
                [drive.drive_letter, drive.label, drive.auto_mount, drive.auto_unmount],
            )
        self.settings.endArray()
        self.settings.endGroup()

    def get_proximity(self) -> int:
        """
        Validates preference value proxmity_seconds against standard list.

        Given the user could enter any old value into the preferences, need to validate
        it. The validation technique is to match whatever value is in the preferences
        with the closest value we need, which is found in the list of int
        proximity_time_steps.

        For the algorithm, see:
        http://stackoverflow.com/questions/12141150/from-list-of-integers-get-number-closest-to-a-given-value
        No need to use bisect list, as our list is tiny, and using min has the advantage
        of getting the closest value.

        Note: we store the value in seconds, but use it in minutes, just in case a user
        one day makes a compelling case to be able to specify a proximity value less
        than 1 minute.

        :return: closest valid value in minutes
        """

        minutes = self.proximity_seconds // 60
        return min(constants.proximity_time_steps, key=lambda x: abs(x - minutes))

    def set_proximity(self, minutes: int) -> None:
        self.proximity_seconds = minutes * 60

    def _pref_list_uses_component(
        self, pref_list, pref_component, offset: int = 1
    ) -> bool:
        for i in range(0, len(pref_list), 3):
            if pref_list[i + offset] == pref_component:
                return True
        return False

    def any_pref_uses_stored_sequence_no(self) -> bool:
        """
        :return True if any of the pref lists contain a stored sequence no
        """
        for pref_list in self.get_pref_lists(file_name_only=True):
            if self._pref_list_uses_component(pref_list, STORED_SEQ_NUMBER):
                return True
        return False

    def any_pref_uses_session_sequence_no(self) -> bool:
        """
        :return True if any of the pref lists contain a session sequence no
        """
        for pref_list in self.get_pref_lists(file_name_only=True):
            if self._pref_list_uses_component(pref_list, SESSION_SEQ_NUMBER):
                return True
        return False

    def any_pref_uses_sequence_letter_value(self) -> bool:
        """
        :return True if any of the pref lists contain a sequence letter
        """
        for pref_list in self.get_pref_lists(file_name_only=True):
            if self._pref_list_uses_component(pref_list, SEQUENCE_LETTER):
                return True
        return False

    def photo_rename_pref_uses_downloads_today(self) -> bool:
        """
        :return: True if the photo rename pref list contains a downloads today
        """
        return self._pref_list_uses_component(self.photo_rename, DOWNLOAD_SEQ_NUMBER)

    def video_rename_pref_uses_downloads_today(self) -> bool:
        """
        :return: True if the video rename pref list contains a downloads today
        """
        return self._pref_list_uses_component(self.video_rename, DOWNLOAD_SEQ_NUMBER)

    def photo_rename_pref_uses_stored_sequence_no(self) -> bool:
        """
        :return: True if the photo rename pref list contains a stored sequence no
        """
        return self._pref_list_uses_component(self.photo_rename, STORED_SEQ_NUMBER)

    def video_rename_pref_uses_stored_sequence_no(self) -> bool:
        """
        :return: True if the video rename pref list contains a stored sequence no
        """
        return self._pref_list_uses_component(self.video_rename, STORED_SEQ_NUMBER)

    def check_prefs_for_validity(self) -> tuple[bool, str]:
        """
        Checks photo & video rename, and subfolder generation
        preferences ensure they follow name generation rules. Moreover,
        subfolder name specifications must not:
        1. start with a separator
        2. end with a separator
        3. have two separators in a row

        :return: tuple with two values: (1) bool and error message if
         prefs are invalid (else empty string)
        """

        msg = ""
        valid = True
        tests = (
            (self.photo_rename, DICT_IMAGE_RENAME_L0),
            (self.video_rename, DICT_VIDEO_RENAME_L0),
            (self.photo_subfolder, DICT_SUBFOLDER_L0),
            (self.video_subfolder, DICT_VIDEO_SUBFOLDER_L0),
        )

        # test file renaming
        for pref, pref_defn in tests[:2]:
            try:
                check_pref_valid(pref_defn, pref)
            except PrefError as e:
                valid = False
                msg += e.msg + "\n"

        # test subfolder generation
        for pref, pref_defn in tests[2:]:
            try:
                check_pref_valid(pref_defn, pref)

                L1s = [pref[i] for i in range(0, len(pref), 3)]

                if L1s[0] == SEPARATOR:
                    raise PrefValueKeyComboError(
                        _("Subfolder preferences should not start with a %s") % os.sep
                    )
                elif L1s[-1] == SEPARATOR:
                    raise PrefValueKeyComboError(
                        _("Subfolder preferences should not end with a %s") % os.sep
                    )
                else:
                    for i in range(len(L1s) - 1):
                        if L1s[i] == SEPARATOR and L1s[i + 1] == SEPARATOR:
                            raise PrefValueKeyComboError(
                                _(
                                    "Subfolder preferences should not contain two %s "
                                    "one after the other"
                                )
                                % os.sep
                            )

            except PrefError as e:
                valid = False
                msg += e.msg + "\n"

        return valid, msg

    def _filter_duplicate_generation_prefs(self, preset_type: PresetPrefType) -> None:
        preset_names, preset_pref_lists = self.get_custom_presets(
            preset_type=preset_type
        )
        seen = set()
        filtered_names = []
        filtered_pref_lists = []
        duplicates = []
        for name, pref_list in zip(preset_names, preset_pref_lists):
            value = tuple(pref_list)
            if value in seen:
                duplicates.append(name)
            else:
                seen.add(value)
                filtered_names.append(name)
                filtered_pref_lists.append(pref_list)

        if duplicates:
            human_readable = preset_type.name[len("preset_") :].replace("_", " ")
            logging.warning(
                "Removed %s duplicate(s) from %s presets: %s",
                len(duplicates),
                human_readable,
                make_internationalized_list(duplicates),
            )
            self.set_custom_presets(
                preset_type=preset_type,
                preset_names=filtered_names,
                preset_pref_lists=filtered_pref_lists,
            )

    def filter_duplicate_generation_prefs(self) -> None:
        """
        Remove any duplicate subfolder generation or file renaming custom presets
        """

        logging.info("Checking for duplicate name generation preference values")
        for preset_type in PresetPrefType:
            self._filter_duplicate_generation_prefs(preset_type)

    def must_synchronize_raw_jpg(self) -> bool:
        """
        :return: True if synchronize_raw_jpg is True and photo
        renaming uses sequence values
        """
        if self.synchronize_raw_jpg:
            for s in LIST_SEQUENCE_L1:
                if self._pref_list_uses_component(self.photo_rename, s, 1):
                    return True
        return False

    def format_pref_list_for_pretty_print(self, pref_list) -> str:
        """
        :return: string useful for printing the preferences
        """

        v = ""
        for i in range(0, len(pref_list), 3):
            c = ":" if pref_list[i + 1] or pref_list[i + 2] else ""
            s = f"{pref_list[i]}{c} "

            if pref_list[i + 1]:
                s = f"{s}{pref_list[i + 1]}"
            if pref_list[i + 2]:
                s = f"{s} ({pref_list[i + 2]})"
            v += s + "\n"
        return v

    def get_pref_lists(self, file_name_only: bool) -> tuple[list[str], ...]:
        """
        :return: a tuple of the photo & video rename and subfolder
         generation preferences
        """
        if file_name_only:
            return self.photo_rename, self.video_rename
        else:
            return (
                self.photo_rename,
                self.photo_subfolder,
                self.video_rename,
                self.video_subfolder,
            )

    def get_day_start_qtime(self) -> QTime:
        """
        :return: day start time in QTime format, resetting to midnight on value error
        """
        try:
            h, m = self.day_start.split(":")
            h = int(h)
            m = int(m)
            assert 0 <= h <= 23
            assert 0 <= m <= 59
            return QTime(h, m)
        except (ValueError, AssertionError):
            logging.error(
                "'Start of day' preference value %s is corrupted. Resetting to "
                "midnight.",
                self.day_start,
            )
            self.day_start = "0:0"
            return QTime(0, 0)

    def get_checkable_value(self, key: str) -> Qt.CheckState:
        """
        Gets a boolean preference value using Qt's CheckState values
        :param key: the preference item to get
        :return: value converted from bool to an Qt.CheckState enum value
        """

        value = self[key]
        if value:
            return Qt.Checked
        else:
            return Qt.Unchecked

    def pref_uses_job_code(self, pref_list: list[str]) -> bool:
        """Returns True if the particular preference contains a job code"""
        return any(pref_list[i] == JOB_CODE for i in range(0, len(pref_list), 3))

    def any_pref_uses_job_code(self) -> bool:
        """Returns True if any of the preferences contain a job code"""
        for pref_list in self.get_pref_lists(file_name_only=False):
            if self.pref_uses_job_code(pref_list):
                return True
        return False

    def file_type_uses_job_code(self, file_type: FileType) -> bool:
        """
        Returns True if either the subfolder generation or file
        renaming for the file type uses a Job Code.
        """

        if file_type == FileType.photo:
            pref_lists = self.photo_rename, self.photo_subfolder
        else:
            pref_lists = self.video_rename, self.video_subfolder

        return any(self.pref_uses_job_code(pref_list) for pref_list in pref_lists)

    def most_recent_job_code(self, missing: str | None = None) -> str:
        """
        Get the most recent Job Code used (which is assumed to be at the top).
        :param missing: If there is no Job Code, and return this default value
        :return: most recent job code, or missing, or if not found, ''
        """

        if len(self.job_codes) > 0:
            value = self.job_codes[0]
            return value or missing or ""
        elif missing is not None:
            return missing
        else:
            return ""

    def photo_rename_index(self, preset_pref_lists: list[list[str]]) -> int:
        """
        Matches the photo pref list with program filename generation
        defaults and the user's presets.

        :param preset_pref_lists: list of custom presets
        :return: -1 if no match (i.e. custom), or the index into
         PHOTO_RENAME_MENU_DEFAULTS_CONV + photo rename presets if it matches
        """

        rename = PHOTO_RENAME_MENU_DEFAULTS_CONV + tuple(preset_pref_lists)
        try:
            return rename.index(self.photo_rename)
        except ValueError:
            return -1

    def video_rename_index(self, preset_pref_lists: list[list[str]]) -> int:
        """
        Matches the video pref list with program filename generation
        defaults and the user's presets.

        :param preset_pref_lists: list of custom presets
        :return: -1 if no match (i.e. custom), or the index into
         VIDEO_RENAME_MENU_DEFAULTS_CONV + video rename presets if it matches
        """

        rename = VIDEO_RENAME_MENU_DEFAULTS_CONV + tuple(preset_pref_lists)
        try:
            return rename.index(self.video_rename)
        except ValueError:
            return -1

    def add_list_value(self, key, value, max_list_size=0) -> None:
        """
        Add value to pref list if it doesn't already exist.

        Values are added to the start of the list.

        An empty list contains only one item: ['']

        :param key: the preference key
        :param value: the value to add
        :param max_list_size: if non-zero, the list's last value will be deleted
        """

        if len(self[key]) == 1 and self[key][0] == "":
            self[key] = [value]
        elif value not in self[key]:
            # Must assign the value like this, otherwise the preference value
            # will not be updated:
            if max_list_size:
                self[key] = [value] + self[key][: max_list_size - 1]
            else:
                self[key] = [value] + self[key]

    def del_list_value(self, key: str, value) -> None:
        """
        Remove a value from the pref list indicated by key.

        Exceptions are not caught.

        An empty list contains only one item: ['']

        :param key: the preference key
        :param value: the value to delete
        """

        # Must remove the value like this, otherwise the preference value
        # will not be updated:
        pref_value = self[key]
        pref_value.remove(value)
        self[key] = pref_value

        if len(self[key]) == 0:
            self[key] = [""]

    def list_not_empty(self, key: str) -> bool:
        """
        In our pref schema, an empty list is [''], not []

        :param key: the preference value to examine
        :return: True if the pref list is not empty
        """

        return bool(self[key] and self[key][0])

    def reset(self) -> None:
        """
        Reset all program preferences to their default settings
        """
        self.settings.clear()
        self.program_version = raphodo.__about__.__version__

    def upgrade_prefs(self, previous_version: Version) -> None:
        """
        Upgrade the user's preferences if needed.

        :param previous_version: previous version of
        Rapid Photo Downloader
        """

        photo_video_rename_change = parse("0.9.0a4")
        if previous_version < photo_video_rename_change:
            for key in ("photo_rename", "video_rename"):
                pref_list, case = upgrade_pre090a4_rename_pref(self[key])
                if pref_list != self[key]:
                    self[key] = pref_list
                    logging.info("Upgraded %s preference value", key.replace("_", " "))
                if case is not None:
                    if key == "photo_rename":
                        self.photo_extension = case
                    else:
                        self.video_extension = case

        v090a5 = parse("0.9.0a5")
        if previous_version < v090a5:
            # Versions prior to 0.9.0a5 incorrectly set the conflict resolution value
            # when importing preferences from 0.4.11 or earlier
            try:
                value = self.conflict_resolution
            except TypeError:
                self.settings.endGroup()
                default = self.defaults["conflict_resolution"]
                default_name = constants.ConflictResolution(default).name
                logging.warning(
                    "Resetting Conflict Resolution preference value to %s", default_name
                )
                self.conflict_resolution = default
            # destinationButtonPressed is no longer used by 0.9.0a5
            self.settings.beginGroup("MainWindow")
            key = "destinationButtonPressed"
            try:
                if self.settings.contains(key):
                    logging.debug("Removing preference value %s", key)
                    self.settings.remove(key)
            except Exception:
                logging.warning("Unknown error removing %s preference value", key)
            self.settings.endGroup()

        v090b6 = parse("0.9.0b6")
        key = "warn_broken_or_missing_libraries"
        group = "Display"
        if previous_version < v090b6 and not self.value_is_set(key, group):
            # Versions prior to 0.9.0b6 may have a preference value
            # 'warn_no_libmediainfo' which is now renamed to
            # 'broken_or_missing_libraries'
            if self.value_is_set("warn_no_libmediainfo", group):
                self.settings.beginGroup(group)
                v = self.settings.value("warn_no_libmediainfo", True, bool)
                self.settings.remove("warn_no_libmediainfo")
                self.settings.endGroup()
                logging.debug(
                    "Transferring preference value %s for warn_no_libmediainfo to "
                    "warn_broken_or_missing_libraries",
                    v,
                )
                self.warn_broken_or_missing_libraries = v
            else:
                logging.debug(
                    "Not transferring preference value warn_no_libmediainfo to "
                    "warn_broken_or_missing_libraries because it doesn't exist"
                )

        v093a1 = parse("0.9.3a1")
        key = "scan_specific_folders"
        group = "Device"
        if previous_version < v093a1 and not self.value_is_set(key, group):
            # Versions prior to 0.9.3a1 used a preference value to indicate if
            # devices lacking a DCIM folder should be scanned. It is now renamed
            # to 'scan_specific_folders'
            if self.value_is_set("device_without_dcim_autodetection"):
                self.settings.beginGroup(group)
                v = self.settings.value("device_without_dcim_autodetection", True, bool)
                self.settings.remove("device_without_dcim_autodetection")
                self.settings.endGroup()
                self.settings.endGroup()
                logging.debug(
                    "Transferring preference value %s for "
                    "device_without_dcim_autodetection to scan_specific_folders as %s",
                    v,
                    not v,
                )
                self.scan_specific_folders = not v
            else:
                logging.debug(
                    "Not transferring preference value "
                    "device_without_dcim_autodetection to scan_specific_folders "
                    "because it doesn't exist"
                )

        v0919b2 = parse("0.9.19b2")
        key = "ignored_paths"
        group = "Device"
        if previous_version < v0919b2 and self.value_is_set(key, group):
            # Versions prior to 0.9.19b2 did not include all the ignored paths
            # introduced in 0.9.16 and 0.9.19b2. If the user already has some
            # values, these new defaults will not be added automatically. So add
            # them here.
            for value in ("THMBNL", "__MACOSX"):
                # If the value is not already in the list, add it
                logging.info("Adding folder '%s' to list of ignored paths" % value)
                self.add_list_value(key=key, value=value)

        v0927a3 = parse("0.9.27a3")
        if previous_version < v0927a3 and self.value_is_set(key, group):
            # Versions prior to 0.9.27a3 did not include all the ignored paths
            # included in that version
            logging.info("Adding folder 'Screenshots' to list of ignored paths")
            self.add_list_value(key=key, value="Screenshots")

    def check_show_system_folders(self) -> None:
        """
        Determine if system folders should be shown because a download source
        or destination is on a system path, i.e. not in /home, /media, /mnt, or
        possibly /run

        Adjusts show_system_folders setting to True if necessary.
        """

        if self.show_system_folders:
            return

        if self.source_or_destination_is_system_folder():
            logging.info(
                "Forcibly setting show system folders to true",
            )
            self.show_system_folders = True

    def source_or_destination_is_system_folder(self) -> bool:
        """
        :return: True if the "This Computer", photo or video destination is
        on a system folder
        """

        system_dir_located = False
        non_system_root_folders = constants.non_system_root_folders
        if get_media_dir().startswith("/run"):
            non_system_root_folders.append("/run")
        for path, name in (
            (self.photo_download_folder, "Photo download folder"),
            (self.video_download_folder, "Video download folder"),
            (self.this_computer_path, "This computer path"),
        ):
            parts = Path(path).resolve().parts
            if path and (
                len(parts) < 2 or f"/{parts[1]}" not in non_system_root_folders
            ):
                logging.debug("'%s' %s is a system directory", name, path)
                system_dir_located = True
                break
        return system_dir_located

    def validate_max_CPU_cores(self) -> None:
        logging.debug("Validating CPU core count for thumbnail generation...")
        available = available_cpu_count(physical_only=True)
        logging.debug("...%s physical cores detected", available)
        if self.max_cpu_cores > available:
            logging.info("Setting CPU Cores for thumbnail generation to %s", available)
            self.max_cpu_cores = available

    def validate_ignore_unhandled_file_exts(self) -> None:
        # logging.debug('Validating list of file extension to not warn about...')
        self.ignore_unhandled_file_exts = [
            ext.upper()
            for ext in self.ignore_unhandled_file_exts
            if ext.lower() not in ALL_KNOWN_EXTENSIONS
        ]

    def warn_about_unknown_file(self, ext: str) -> bool:
        if not self.warn_unhandled_files:
            return False

        if not self.ignore_unhandled_file_exts[0]:
            return True

        return ext.upper() not in self.ignore_unhandled_file_exts

    def settings_path(self) -> str:
        """
        :return: the full path of the settings file
        """
        return self.settings.fileName()


def match_pref_list(pref_lists: list[list[str]], user_pref_list: list[str]) -> int:
    try:
        return pref_lists.index(user_pref_list)
    except ValueError:
        return -1



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

import math

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import (
    QFont,  # noqa: F401
    QFontMetrics,
    QGuiApplication,
    QIcon,
    QPainter,
    QPaintEvent,
)
from PyQt5.QtWidgets import QApplication, QPushButton, QSizePolicy

from raphodo.ui.rotatedpushbutton import FlatButton
from raphodo.ui.viewutils import darkModeIcon, is_dark_mode


class TopPushButton(QPushButton, FlatButton):
    def __init__(self, text, parent, extra_top: int = 0) -> None:
        """

        :param text: text to display in the button
        :param extra_top: extra spacing at the top of the widget
        :param parent: parent widget
        """

        super().__init__(text, parent)
        self.rapidApp = parent
        self.setCheckable(True)
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        font = self.font()  # type: QFont
        top_row_font_size = font.pointSize() + 8
        self.top_row_icon_size = top_row_font_size + 10
        font.setPointSize(top_row_font_size)
        self.setFont(font)

        font_height = QFontMetrics(font).height()
        self.padding_side = math.ceil(font_height / 3.5)
        padding_bottom = math.ceil(font_height / 5.6)
        padding_top = padding_bottom + extra_top

        self.non_elided_text = ""

        padding = (
            f"padding-left: {self.padding_side}px; "
            f"padding-right: {self.padding_side}px; "
            f"padding-top: {padding_top}px; "
            f"padding-bottom: {padding_bottom}px; "
        )
        self.setFlatStyle(self, darker_if_checked=False, padding=padding)

    def text(self) -> str:
        return self.non_elided_text

    def setText(self, text: str) -> None:
        self.non_elided_text = text
        self.update()

    def setIcon(self, icon: QIcon) -> None:
        size = QSize(self.top_row_icon_size, self.top_row_icon_size)
        icon = darkModeIcon(icon=icon, size=size)
        super().setIcon(icon)
        self.setIconSize(size)

    def paintEvent(self, event: QPaintEvent):
        """
        Override default rendering to elide button text if it is bigger than half the
        window size
        """

        painter = QPainter(self)
        metrics = painter.fontMetrics()
        right_element_widths = (
            self.rapidApp.downloadButton.width() + self.rapidApp.menuButton.width()
        )
        window_width = self.rapidApp.width()
        window_half = window_width / 2
        if right_element_widths > window_half:
            maximum_width = window_width - right_element_widths
        else:
            maximum_width = window_half
        maximum_width -= self.padding_side - self.top_row_icon_size

        # account for situations where we might have negative values, i.e., display some
        # text at least
        maximum_width = max(30, maximum_width)

        usable_width = round(0.9 * maximum_width)
        elided_text = metrics.elidedText(
            self.non_elided_text, Qt.ElideMiddle, usable_width
        )
        super().setText(elided_text)
        super().paintEvent(event)


def DownloadButtonHeight() -> tuple[int, int]:
    font_height = (
        QFontMetrics(QApplication.font())
        .tightBoundingRect(_("Download 8 Photos and 10 Videos"))
        .height()
    )
    padding = math.ceil(font_height * 1.7)
    height = font_height // 2 * 6
    return height, padding


class DownloadButton(QPushButton):
    """
    Button used to initiate downloads
    """

    def __init__(self, text: str, parent) -> None:
        super().__init__(text, parent)

        self.rapidApp = parent
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        height, padding = DownloadButtonHeight()
        radius = height // 2

        palette = QGuiApplication.palette()
        primaryColor = palette.highlight().color()
        borderColor = primaryColor.darker(105)
        hoverColor = palette.highlight().color().darker(106)
        hoverBorderColor = hoverColor.darker(105)
        primaryTextColor = palette.highlightedText().color()

        if is_dark_mode():
            disabledColor = palette.window().color().lighter(130)
            disabledBorderColor = disabledColor.lighter(115)
            disabledTextColor = palette.highlightedText().color()
        else:
            disabledColor = palette.window().color().darker(120)
            disabledBorderColor = disabledColor.darker(105)
            disabledTextColor = palette.highlightedText().color()

        # outline:none is used to remove the rectangle that appears on a
        # button when the button has focus
        # http://stackoverflow.com/questions/17280056/qt-css-decoration-on-focus

        self.setStyleSheet(
            f"""
            QPushButton {{
            background-color: {primaryColor.name()};
            outline: none;
            padding-left: {padding}px;
            padding-right: {padding}px;
            border-radius: {radius}px;
            border: 1px solid {borderColor.name()};
            height: {height}px;
            color: {primaryTextColor.name()};
            }}
            QPushButton:hover {{
            background-color: {hoverColor.name()};
            border: 1px solid {hoverBorderColor.name()};
            }}
            QPushButton:disabled {{
            background-color: {disabledColor.name()};
            color: {disabledTextColor.name()};
            border: 1px solid {disabledBorderColor.name()};
            }}
            """
        )

    def setText(self, text: str) -> None:
        super().setText(text)
        self.rapidApp.sourceButton.updateGeometry()



#!/usr/bin/env python3

# Copyright (C) 2015-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2015-2024, Damon Lynch"

# ruff: noqa: E402

import contextlib
import logging
import os
import pickle
import sys
from collections import namedtuple
from urllib.request import pathname2url

import gi

gi.require_version("Gst", "1.0")
from gi.repository import Gst
from PyQt5.QtCore import QBuffer, QIODevice, QSize, Qt
from PyQt5.QtGui import QImage, QTransform

import raphodo.metadata.exiftool as exiftool
from raphodo.cache import FdoCacheLarge, FdoCacheNormal, ThumbnailCacheSql
from raphodo.constants import (
    ExtractionProcessing,
    ExtractionTask,
    FileType,
    ThumbnailCacheDiskStatus,
    ThumbnailCacheStatus,
    ThumbnailSize,
)
from raphodo.heif import have_heif_module, load_heif
from raphodo.interprocess import (
    GenerateThumbnailsResults,
    LoadBalancerWorker,
    ThumbnailExtractorArgument,
)
from raphodo.rpdfile import Photo, RPDFile, Video
from raphodo.ui.filmstrip import add_filmstrip
from raphodo.utilities import image_large_enough_fdo, show_errors, stdchannel_redirected

have_gst = Gst.init_check(None)


def gst_version() -> str:
    """
    :return: version of gstreamer, if it exists and is functioning, else ''
    """

    if have_gst:
        try:
            return Gst.version_string().replace("GStreamer ", "")
        except Exception:
            pass
    return ""


def get_video_frame(
    full_file_name: str,
    offset: float | None = 5.0,
    caps=Gst.Caps.from_string("image/png"),
) -> bytes | None:
    """
    Source: https://gist.github.com/dplanella/5563018

    :param full_file_name: file and path of the video
    :param offset: how many seconds into the video to read
    :param caps:
    :return: gstreamer buffer
    """

    logging.debug("Using gstreamer to generate thumbnail from %s", full_file_name)
    pipeline = Gst.parse_launch("playbin")
    pipeline.props.uri = f"file://{pathname2url(os.path.abspath(full_file_name))}"
    pipeline.props.audio_sink = Gst.ElementFactory.make("fakesink", "fakeaudio")
    pipeline.props.video_sink = Gst.ElementFactory.make("fakesink", "fakevideo")
    pipeline.set_state(Gst.State.PAUSED)
    # Wait for state change to finish.
    pipeline.get_state(Gst.CLOCK_TIME_NONE)

    # Seek offset .10 seconds into the video as a minimum
    if not offset:
        offset = 0.5 * Gst.SECOND

    # Duration is unreliable because when we are dealing with camera videos,
    # we're only downloading a snapshot, i.e. 2 seconds of a 1 minute video.
    # But no matter what, don't want to exceed it.
    duration = pipeline.query_duration(Gst.Format.TIME)[1]
    offset = min(duration, offset)

    try:
        v = pipeline.seek_simple(
            Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.KEY_UNIT, offset
        )
        assert v
    except AssertionError:
        logging.warning(
            "seek_simple() failed for %s. Is the necessary gstreamer plugin installed "
            "for this file format?",
            full_file_name,
        )
        return None
    # Wait for seek to finish.
    pipeline.get_state(Gst.CLOCK_TIME_NONE)  # alternative is Gst.SECOND * 10
    sample = pipeline.emit("convert-sample", caps)
    if sample is not None:
        buffer = sample.get_buffer()
        pipeline.set_state(Gst.State.NULL)
        return buffer.extract_dup(0, buffer.get_size())
    else:
        return None


PhotoDetails = namedtuple("PhotoDetails", "thumbnail, orientation")


def qimage_to_png_buffer(image: QImage) -> QBuffer:
    """
    Save the image data in PNG format in a QBuffer, whose data can then
    be extracted using the data() member function.
    :param image: the image to be converted
    :return: the buffer
    """

    buffer = QBuffer()
    buffer.open(QIODevice.WriteOnly)
    # Quality 100 means uncompressed.
    image.save(buffer, "PNG", quality=100)
    return buffer


def crop_160x120_thumbnail(thumbnail: QImage, vertical_space: int = 8) -> QImage:
    """
    Remove black bands from the top and bottom of thumbnail
    :param thumbnail: thumbnail to crop
    :param vertical_space: how much to remove from the top and bottom
    :return: cropped thumbnail
    """
    if thumbnail.width() == 160 and thumbnail.height() == 120:
        return thumbnail.copy(0, vertical_space, 160, 120 - vertical_space * 2)
    elif thumbnail.width() == 120 and thumbnail.height() == 160:
        return thumbnail.copy(vertical_space, 0, 120 - vertical_space * 2, 160)
    else:
        return thumbnail


class ThumbnailExtractor(LoadBalancerWorker):
    # Exif rotation constants
    rotate_0 = "1"
    rotate_90 = "6"
    rotate_180 = "3"
    rotate_270 = "8"

    maxStandardSize = QSize(
        max(ThumbnailSize.width, ThumbnailSize.height),
        max(ThumbnailSize.width, ThumbnailSize.height),
    )

    def __init__(self) -> None:
        self.thumbnailSizeNeeded = QSize(ThumbnailSize.width, ThumbnailSize.height)
        self.thumbnail_cache = ThumbnailCacheSql(create_table_if_not_exists=False)
        self.fdo_cache_large = FdoCacheLarge()
        self.fdo_cache_normal = FdoCacheNormal()

        super().__init__("Thumbnail Extractor")

    def rotate_thumb(self, thumbnail: QImage, orientation: str) -> QImage:
        """
        If required return a rotated copy the thumbnail
        :param thumbnail: thumbnail to rotate
        :param orientation: EXIF orientation tag
        :return: possibly rotated thumbnail
        """

        if orientation == self.rotate_90:
            thumbnail = thumbnail.transformed(QTransform().rotate(90))
        elif orientation == self.rotate_270:
            thumbnail = thumbnail.transformed(QTransform().rotate(270))
        elif orientation == self.rotate_180:
            thumbnail = thumbnail.transformed(QTransform().rotate(180))
        return thumbnail

    def image_large_enough(self, size: QSize) -> bool:
        """Check if image is equal or bigger than thumbnail size."""

        return (
            size.width() >= self.thumbnailSizeNeeded.width()
            or size.height() >= self.thumbnailSizeNeeded.height()
        )

    def _extract_256_thumb(
        self,
        rpd_file: RPDFile,
        processing: set[ExtractionProcessing],
        orientation: str | None,
    ) -> PhotoDetails:
        thumbnail = None
        data = rpd_file.metadata.get_preview_256()
        if isinstance(data, bytes):
            thumbnail = QImage.fromData(data)
            if thumbnail.isNull():
                thumbnail = None
            else:
                if thumbnail.width() > 160 or thumbnail.height() > 120:
                    processing.add(ExtractionProcessing.resize)

        return PhotoDetails(thumbnail, orientation)

    def _extract_metadata(
        self, rpd_file: RPDFile, processing: set[ExtractionProcessing]
    ) -> PhotoDetails:
        thumbnail = orientation = None
        with contextlib.suppress(Exception):
            orientation = rpd_file.metadata.orientation()

        rpd_file.mdatatime = rpd_file.metadata.timestamp(missing=0.0)

        # Not all files have an exif preview, but some do
        # (typically CR2, ARW, PEF, RW2).
        # If they exist, they are (almost!) always 160x120

        # TODO how about thumbnail_cache_status?
        if self.write_fdo_thumbnail and rpd_file.fdo_thumbnail_256 is None:
            photo_details = self._extract_256_thumb(
                rpd_file=rpd_file, processing=processing, orientation=orientation
            )
            if photo_details.thumbnail is not None:
                return photo_details
            # if no valid preview found, fall back to the code below and make do with
            # the best we can get

        preview = rpd_file.metadata.get_small_thumbnail_or_first_indexed_preview()
        if preview:
            thumbnail = QImage.fromData(preview)
            if thumbnail.isNull():
                thumbnail = None
            else:
                if thumbnail.width() < thumbnail.height() and orientation in (
                    self.rotate_270,
                    self.rotate_90,
                ):
                    # The orientation has already been applied to the thumbnail
                    logging.debug(
                        "Already rotated: %s", rpd_file.get_current_full_file_name()
                    )
                    orientation = self.rotate_0

                if max(thumbnail.width(), thumbnail.height()) > 160:
                    logging.debug("Resizing: %s", rpd_file.get_current_full_file_name())
                    processing.add(ExtractionProcessing.resize)
                elif not rpd_file.is_jpeg():
                    processing.add(ExtractionProcessing.strip_bars_photo)

        return PhotoDetails(thumbnail, orientation)

    def get_disk_photo_thumb(
        self,
        rpd_file: Photo,
        full_file_name: str,
        processing: set[ExtractionProcessing],
        force_exiftool: bool,
    ) -> PhotoDetails:
        """
        Get the photo's thumbnail from a file that is on disk.

        Sets rpd_file's mdatatime.

        :param rpd_file: file details
        :param full_file_name: full name of the file from which to get the metadata
        :param processing: processing extraction tasks to complete,
        :param force_exiftool: whether to force the use of ExifTool to load the metadata
        :return: thumbnail and its orientation
        """

        orientation = None
        thumbnail = None
        photo_details = PhotoDetails(thumbnail, orientation)
        if rpd_file.load_metadata(
            full_file_name=full_file_name,
            et_process=self.exiftool_process,
            force_exiftool=force_exiftool,
        ):
            photo_details = self._extract_metadata(rpd_file, processing)
            thumbnail = photo_details.thumbnail

        if thumbnail is not None:
            return photo_details

        if rpd_file.is_loadable():
            thumbnail = QImage(full_file_name)
            processing.add(ExtractionProcessing.resize)
            if not rpd_file.from_camera:
                processing.remove(ExtractionProcessing.orient)
            if thumbnail.isNull():
                thumbnail = None
                logging.warning(
                    f"Unable to create a thumbnail out of the file: {full_file_name}"
                )

        return PhotoDetails(thumbnail, orientation)

    def get_from_buffer(
        self,
        rpd_file: Photo,
        raw_bytes: bytearray,
        processing: set[ExtractionProcessing],
    ) -> PhotoDetails:
        if not rpd_file.load_metadata(
            raw_bytes=raw_bytes, et_process=self.exiftool_process
        ):
            return PhotoDetails(None, None)
        else:
            return self._extract_metadata(rpd_file, processing)

    def get_photo_orientation(
        self,
        rpd_file: Photo,
        force_exiftool: bool,
        full_file_name: str | None = None,
        raw_bytes: bytearray | None = None,
    ) -> str | None:
        if rpd_file.metadata is None:
            self.load_photo_metadata(
                rpd_file=rpd_file,
                full_file_name=full_file_name,
                raw_bytes=raw_bytes,
                force_exiftool=force_exiftool,
            )

        if rpd_file.metadata is not None:
            try:
                return rpd_file.metadata.orientation()
            except Exception:
                pass
        return None

    def assign_mdatatime(
        self,
        rpd_file: Photo | Video,
        force_exiftool: bool,
        full_file_name: str | None = None,
        raw_bytes: bytearray | None = None,
    ) -> None:
        """
        Load the file's metadata and assign the metadata time to the rpd file
        """

        if rpd_file.file_type == FileType.photo:
            self.assign_photo_mdatatime(
                rpd_file=rpd_file,
                full_file_name=full_file_name,
                raw_bytes=raw_bytes,
                force_exiftool=force_exiftool,
            )
        else:
            self.assign_video_mdatatime(
                rpd_file=rpd_file, full_file_name=full_file_name
            )

    def assign_photo_mdatatime(
        self,
        rpd_file: Photo,
        force_exiftool: bool,
        full_file_name: str | None = None,
        raw_bytes: bytearray | None = None,
    ) -> None:
        """
        Load the photo's metadata and assign the metadata time to the rpd file
        """

        self.load_photo_metadata(
            rpd_file=rpd_file,
            full_file_name=full_file_name,
            raw_bytes=raw_bytes,
            force_exiftool=force_exiftool,
        )
        if rpd_file.metadata is not None and rpd_file.date_time() is None:
            rpd_file.mdatatime = 0.0

    def load_photo_metadata(
        self,
        rpd_file: Photo,
        force_exiftool: bool,
        full_file_name: str | None = None,
        raw_bytes: bytearray | None = None,
    ) -> None:
        """
        Load the photo's metadata into the rpd file
        """

        if raw_bytes is not None:
            if rpd_file.is_jpeg_type():
                rpd_file.load_metadata(
                    app1_segment=raw_bytes, et_process=self.exiftool_process
                )
            else:
                rpd_file.load_metadata(
                    raw_bytes=raw_bytes, et_process=self.exiftool_process
                )
        else:
            rpd_file.load_metadata(
                full_file_name=full_file_name,
                et_process=self.exiftool_process,
                force_exiftool=force_exiftool,
            )

    def assign_video_mdatatime(self, rpd_file: Video, full_file_name: str) -> None:
        """
        Load the video's metadata and assign the metadata time to the rpd file
        """

        if rpd_file.metadata is None:
            rpd_file.load_metadata(
                full_file_name=full_file_name, et_process=self.exiftool_process
            )
        if rpd_file.date_time() is None:
            rpd_file.mdatatime = 0.0

    def get_video_rotation(self, rpd_file: Video, full_file_name: str) -> str | None:
        """
        Some videos have a rotation tag. If this video does, return it.
        """

        if rpd_file.metadata is None:
            rpd_file.load_metadata(
                full_file_name=full_file_name, et_process=self.exiftool_process
            )
        orientation = rpd_file.metadata.rotation(missing=None)
        match orientation:
            case 180:
                return self.rotate_180
            case 90:
                return self.rotate_90
            case 270:
                return self.rotate_270
        return None

    def check_for_stop(self, directive: bytes, content: bytes):
        if directive == b"cmd":
            assert content == b"STOP"
            return True
        return False

    def extract_thumbnail(
        self,
        task: ExtractionTask,
        rpd_file: Photo | Video,
        processing: set[ExtractionProcessing],
        data: ThumbnailExtractorArgument,
    ) -> tuple[QImage | None, str | None]:
        """
        Extract the thumbnail using one of a variety of methods,
        depending on the file

        :param task: extraction task to perform
        :param rpd_file: rpd_file to work on
        :param processing: processing tasks
        :param data: some other processing arguments passed to this process
        :return: thumbnail and its orientation, if found
        """

        orientation = None

        if task == ExtractionTask.load_from_exif:
            thumbnail_details = self.get_disk_photo_thumb(
                rpd_file,
                data.full_file_name_to_work_on,
                processing,
                data.force_exiftool,
            )
            thumbnail = thumbnail_details.thumbnail
            if thumbnail is not None:
                orientation = thumbnail_details.orientation

        elif task in (
            ExtractionTask.load_file_directly,
            ExtractionTask.load_file_and_exif_directly,
            ExtractionTask.load_file_directly_metadata_from_secondary,
        ):
            thumbnail = QImage(data.full_file_name_to_work_on)

            if task == ExtractionTask.load_file_and_exif_directly:
                self.assign_photo_mdatatime(
                    rpd_file=rpd_file,
                    full_file_name=data.full_file_name_to_work_on,
                    force_exiftool=data.force_exiftool,
                )
            elif task == ExtractionTask.load_file_directly_metadata_from_secondary:
                self.assign_mdatatime(
                    rpd_file=rpd_file,
                    full_file_name=data.secondary_full_file_name,
                    force_exiftool=data.force_exiftool,
                )

            if ExtractionProcessing.orient in processing:
                orientation = self.get_photo_orientation(
                    rpd_file=rpd_file,
                    full_file_name=data.full_file_name_to_work_on,
                    force_exiftool=data.force_exiftool,
                )

        elif task in (
            ExtractionTask.load_from_bytes,
            ExtractionTask.load_from_bytes_metadata_from_temp_extract,
        ):
            try:
                assert data.thumbnail_bytes is not None
            except AssertionError:
                logging.error(
                    "Thumbnail bytes not extracted for %s (value is None)",
                    rpd_file.get_current_full_file_name(),
                )
            thumbnail = QImage.fromData(data.thumbnail_bytes)
            if (
                thumbnail.width() > self.thumbnailSizeNeeded.width()
                or thumbnail.height() > self.thumbnailSizeNeeded.height()
            ):
                processing.add(ExtractionProcessing.resize)
                processing.remove(ExtractionProcessing.strip_bars_photo)
            if data.exif_buffer and ExtractionProcessing.orient in processing:
                orientation = self.get_photo_orientation(
                    rpd_file=rpd_file,
                    raw_bytes=data.exif_buffer,
                    force_exiftool=data.force_exiftool,
                )
            if task == ExtractionTask.load_from_bytes_metadata_from_temp_extract:
                self.assign_mdatatime(
                    rpd_file=rpd_file,
                    full_file_name=data.secondary_full_file_name,
                    force_exiftool=data.force_exiftool,
                )
                orientation = rpd_file.metadata.orientation()
                os.remove(data.secondary_full_file_name)
                rpd_file.temp_cache_full_file_chunk = ""

        elif task == ExtractionTask.load_from_exif_buffer:
            thumbnail_details = self.get_from_buffer(
                rpd_file, data.exif_buffer, processing
            )
            thumbnail = thumbnail_details.thumbnail
            if thumbnail is not None:
                orientation = thumbnail_details.orientation

        elif task in (
            ExtractionTask.load_heif_directly,
            ExtractionTask.load_heif_and_exif_directly,
        ):
            assert have_heif_module
            thumbnail = load_heif(
                data.full_file_name_to_work_on, process_name=self.identity.decode()
            )

            if task == ExtractionTask.load_heif_and_exif_directly:
                self.assign_photo_mdatatime(
                    rpd_file=rpd_file,
                    full_file_name=data.full_file_name_to_work_on,
                    force_exiftool=data.force_exiftool,
                )
            if ExtractionProcessing.orient in processing:
                orientation = self.get_photo_orientation(
                    rpd_file=rpd_file,
                    full_file_name=data.full_file_name_to_work_on,
                    force_exiftool=data.force_exiftool,
                )

        else:
            assert task in (
                ExtractionTask.extract_from_file,
                ExtractionTask.extract_from_file_and_load_metadata,
            )
            if rpd_file.file_type == FileType.photo:
                self.assign_photo_mdatatime(
                    rpd_file=rpd_file,
                    full_file_name=data.full_file_name_to_work_on,
                    force_exiftool=data.force_exiftool,
                )
                thumbnail_bytes = (
                    rpd_file.metadata.get_small_thumbnail_or_first_indexed_preview()
                )
                if thumbnail_bytes:
                    thumbnail = QImage.fromData(thumbnail_bytes)
                    orientation = rpd_file.metadata.orientation()
            else:
                assert rpd_file.file_type == FileType.video

                if ExtractionTask.extract_from_file_and_load_metadata:
                    self.assign_video_mdatatime(
                        rpd_file=rpd_file, full_file_name=data.full_file_name_to_work_on
                    )
                if not have_gst:
                    thumbnail = None
                else:
                    png = get_video_frame(data.full_file_name_to_work_on, 1.0)
                    if not png:
                        thumbnail = None
                        logging.warning(
                            "Could not extract video thumbnail from %s",
                            data.rpd_file.get_display_full_name(),
                        )
                    else:
                        thumbnail = QImage.fromData(png)
                        if thumbnail.isNull():
                            thumbnail = None
                        else:
                            processing.add(ExtractionProcessing.add_film_strip)
                            orientation = self.get_video_rotation(
                                rpd_file, data.full_file_name_to_work_on
                            )
                            if orientation is not None:
                                processing.add(ExtractionProcessing.orient)
                            processing.add(ExtractionProcessing.resize)

        return thumbnail, orientation

    def process_files(self):
        """
        Loop continuously processing photo and video thumbnails
        """

        logging.debug(f"{self.requester.identity.decode()} worker started")

        while True:
            directive, content = self.requester.recv_multipart()
            if self.check_for_stop(directive, content):
                break

            data = pickle.loads(content)  # type: ThumbnailExtractorArgument

            thumbnail_256 = png_data = None
            task = data.task
            processing = data.processing
            rpd_file = data.rpd_file

            logging.debug(
                "Working on task %s for %s",
                task.name,
                rpd_file.download_name or rpd_file.name,
            )

            self.write_fdo_thumbnail = data.write_fdo_thumbnail

            try:
                if rpd_file.fdo_thumbnail_256 is not None and data.write_fdo_thumbnail:
                    if rpd_file.thumbnail_status != ThumbnailCacheStatus.fdo_256_ready:
                        logging.error(
                            "Unexpected thumbnail cache status for %s: %s",
                            rpd_file.full_file_name,
                            rpd_file.thumbnail_status.name,
                        )
                    thumbnail = thumbnail_256 = QImage.fromData(
                        rpd_file.fdo_thumbnail_256
                    )
                    orientation_unknown = False
                else:
                    thumbnail, orientation = self.extract_thumbnail(
                        task, rpd_file, processing, data
                    )
                    if data.file_to_work_on_is_temporary:
                        os.remove(data.full_file_name_to_work_on)
                        rpd_file.temp_cache_full_file_chunk = ""

                    if thumbnail is not None:
                        if ExtractionProcessing.strip_bars_photo in processing:
                            thumbnail = crop_160x120_thumbnail(thumbnail)
                        elif ExtractionProcessing.strip_bars_video in processing:
                            thumbnail = crop_160x120_thumbnail(thumbnail, 15)
                        if ExtractionProcessing.resize in processing:
                            # Resize the thumbnail before rotating
                            if (
                                orientation == "1" or orientation is None
                            ) and thumbnail.height() > thumbnail.width():
                                # Special case: pictures from some cellphones have
                                # already been rotated
                                thumbnail = thumbnail.scaled(
                                    self.maxStandardSize,
                                    Qt.KeepAspectRatio,
                                    Qt.SmoothTransformation,
                                )
                            else:
                                if (
                                    rpd_file.should_write_fdo()
                                    and image_large_enough_fdo(thumbnail.size())
                                    and max(thumbnail.height(), thumbnail.width()) > 256
                                ):
                                    thumbnail_256 = thumbnail.scaled(
                                        QSize(256, 256),
                                        Qt.KeepAspectRatio,
                                        Qt.SmoothTransformation,
                                    )
                                    thumbnail = thumbnail_256
                                if data.send_thumb_to_main:
                                    thumbnail = thumbnail.scaled(
                                        self.thumbnailSizeNeeded,
                                        Qt.KeepAspectRatio,
                                        Qt.SmoothTransformation,
                                    )
                                else:
                                    thumbnail = None

                            if thumbnail is not None and thumbnail.isNull():
                                thumbnail = None

                    if orientation is not None:
                        if thumbnail is not None:
                            thumbnail = self.rotate_thumb(thumbnail, orientation)
                        if thumbnail_256 is not None:
                            thumbnail_256 = self.rotate_thumb(
                                thumbnail_256, orientation
                            )

                    if ExtractionProcessing.add_film_strip in processing:
                        if thumbnail is not None:
                            thumbnail = add_filmstrip(thumbnail)
                        if thumbnail_256 is not None:
                            thumbnail = add_filmstrip(thumbnail_256)

                    if thumbnail is not None:
                        buffer = qimage_to_png_buffer(thumbnail)
                        png_data = buffer.data()

                    orientation_unknown = (
                        ExtractionProcessing.orient in processing
                        and orientation is None
                    )

                    if (
                        data.send_thumb_to_main
                        and data.use_thumbnail_cache
                        and rpd_file.thumbnail_cache_status
                        == ThumbnailCacheDiskStatus.not_found
                    ):
                        self.thumbnail_cache.save_thumbnail(
                            full_file_name=rpd_file.full_file_name,
                            size=rpd_file.size,
                            mtime=rpd_file.modification_time,
                            mdatatime=rpd_file.mdatatime,
                            generation_failed=thumbnail is None,
                            orientation_unknown=orientation_unknown,
                            thumbnail=thumbnail,
                            camera_model=rpd_file.camera_model,
                        )

                if (
                    thumbnail is not None or thumbnail_256 is not None
                ) and rpd_file.should_write_fdo():
                    if self.write_fdo_thumbnail:
                        # The modification time of the file may have changed when the
                        # file was saved Ideally it shouldn't, but it does sometimes,
                        # e.g. on NTFS! So need to get the modification time from the
                        # saved file.
                        mtime = os.path.getmtime(rpd_file.download_full_file_name)

                        if thumbnail_256 is not None:
                            rpd_file.fdo_thumbnail_256_name = (
                                self.fdo_cache_large.save_thumbnail(
                                    full_file_name=rpd_file.download_full_file_name,
                                    size=rpd_file.size,
                                    modification_time=mtime,
                                    generation_failed=False,
                                    thumbnail=thumbnail_256,
                                    free_desktop_org=False,
                                )
                            )
                            thumbnail_128 = thumbnail_256.scaled(
                                QSize(128, 128),
                                Qt.KeepAspectRatio,
                                Qt.SmoothTransformation,
                            )
                        else:
                            thumbnail_128 = thumbnail.scaled(
                                QSize(128, 128),
                                Qt.KeepAspectRatio,
                                Qt.SmoothTransformation,
                            )
                        rpd_file.fdo_thumbnail_128_name = (
                            self.fdo_cache_normal.save_thumbnail(
                                full_file_name=rpd_file.download_full_file_name,
                                size=rpd_file.size,
                                modification_time=mtime,
                                generation_failed=False,
                                thumbnail=thumbnail_128,
                                free_desktop_org=False,
                            )
                        )
                    elif (
                        thumbnail_256 is not None and rpd_file.fdo_thumbnail_256 is None
                    ):
                        rpd_file.fdo_thumbnail_256 = qimage_to_png_buffer(
                            thumbnail
                        ).data()

                if thumbnail is not None:
                    if orientation_unknown:
                        rpd_file.thumbnail_status = (
                            ThumbnailCacheStatus.orientation_unknown
                        )
                    elif rpd_file.fdo_thumbnail_256 is not None:
                        rpd_file.thumbnail_status = ThumbnailCacheStatus.fdo_256_ready
                    else:
                        rpd_file.thumbnail_status = ThumbnailCacheStatus.ready

            except SystemExit as e:
                self.exiftool_process.terminate()
                sys.exit(e.code)
            except Exception:
                logging.error("Exception working on file %s", rpd_file.full_file_name)
                logging.error("Task: %s", task)
                logging.error("Processing tasks: %s", processing)
                logging.exception("Traceback:")

            # Purge metadata, as it cannot be pickled
            if not data.send_thumb_to_main:
                png_data = None
            rpd_file.metadata = None
            self.sender.send_multipart(
                [
                    b"0",
                    b"data",
                    pickle.dumps(
                        GenerateThumbnailsResults(
                            rpd_file=rpd_file, thumbnail_bytes=png_data
                        ),
                        pickle.HIGHEST_PROTOCOL,
                    ),
                ]
            )
            self.requester.send_multipart([b"", b"", b"OK"])

    def do_work(self):
        if False:
            # exiv2 pumps out a LOT to stderr - use cautiously!
            context = show_errors()
            self.error_stream = sys.stderr
        else:
            # Redirect stderr, hiding error output from exiv2
            context = stdchannel_redirected(sys.stderr, os.devnull)
            self.error_stream = sys.stdout
        with context:
            # In some situations, using a context manager for exiftool can
            # result in exiftool processes not being terminated. So let's
            # handle starting and terminating it manually.
            self.exiftool_process = exiftool.ExifTool()
            self.exiftool_process.start()
            self.process_files()
            self.exit()

    def cleanup_pre_stop(self) -> None:
        logging.debug(
            "Terminating thumbnail extractor ExifTool process for %s",
            self.identity.decode(),
        )
        self.exiftool_process.terminate()


if __name__ == "__main__":
    thumbnail_extractor = ThumbnailExtractor()



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Display file renaming preferences, including sequence numbers
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

import logging

from PyQt5.QtCore import Qt, QTime, pyqtSlot
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFormLayout,
    QGroupBox,
    QLabel,
    QSizePolicy,
    QSpinBox,
    QSplitter,
    QTimeEdit,
    QVBoxLayout,
    QWidget,
)

import raphodo.generatename as gn
import raphodo.metadata.exiftool as exiftool
from raphodo.constants import FileType, NameGenerationType, PresetClass, PresetPrefType
from raphodo.generatenameconfig import (
    DICT_IMAGE_RENAME_L0,
    DICT_VIDEO_RENAME_L0,
    LOWERCASE,
    ORIGINAL_CASE,
    PHOTO_RENAME_MENU_DEFAULTS_CONV,
    UPPERCASE,
    VIDEO_RENAME_MENU_DEFAULTS_CONV,
)
from raphodo.prefs.preferences import DownloadsTodayTracker, Preferences
from raphodo.rpdfile import Photo, Video
from raphodo.ui.nameeditor import PrefDialog, PresetComboBox, make_sample_rpd_file
from raphodo.ui.panelview import QPanelView
from raphodo.ui.viewutils import FlexiFrame, ScrollAreaNoFrame
from raphodo.utilities import platform_c_maxint


class RenameWidget(FlexiFrame):
    """
    Display combo boxes for file renaming and file extension case handling, and
    an example file name
    """

    def __init__(
        self,
        preset_type: PresetPrefType,
        prefs: Preferences,
        exiftool_process: exiftool.ExifTool,
        parent,
    ) -> None:
        super().__init__(parent=parent)
        self.setBackgroundRole(QPalette.Base)
        self.setAutoFillBackground(True)
        self.exiftool_process = exiftool_process
        self.prefs = prefs
        self.preset_type = preset_type
        if preset_type == PresetPrefType.preset_photo_rename:
            self.file_type = FileType.photo
            self.pref_defn = DICT_IMAGE_RENAME_L0
            self.generation_type = NameGenerationType.photo_name
            self.index_lookup = self.prefs.photo_rename_index
            self.pref_conv = PHOTO_RENAME_MENU_DEFAULTS_CONV
            self.generation_type = NameGenerationType.photo_name
        else:
            self.file_type = FileType.video
            self.pref_defn = DICT_VIDEO_RENAME_L0
            self.generation_type = NameGenerationType.video_name
            self.index_lookup = self.prefs.video_rename_index
            self.pref_conv = VIDEO_RENAME_MENU_DEFAULTS_CONV
            self.generation_type = NameGenerationType.video_name

        self.sample_rpd_file = make_sample_rpd_file(
            sample_job_code=self.prefs.most_recent_job_code(missing=_("Job Code")),
            prefs=self.prefs,
            generation_type=self.generation_type,
        )

        layout = QFormLayout()
        self.layout().addLayout(layout)

        self.getCustomPresets()

        self.renameCombo = PresetComboBox(
            prefs=self.prefs,
            preset_names=self.preset_names,
            preset_type=preset_type,
            parent=self,
            edit_mode=False,
        )
        self.setRenameComboIndex()
        self.renameCombo.activated.connect(self.renameComboItemActivated)

        # File extensions
        self.extensionCombo = QComboBox()
        self.extensionCombo.addItem(_(ORIGINAL_CASE), ORIGINAL_CASE)
        self.extensionCombo.addItem(_(UPPERCASE), UPPERCASE)
        self.extensionCombo.addItem(_(LOWERCASE), LOWERCASE)
        if preset_type == PresetPrefType.preset_photo_rename:
            pref_value = self.prefs.photo_extension
        else:
            pref_value = self.prefs.video_extension
        try:
            index = [ORIGINAL_CASE, UPPERCASE, LOWERCASE].index(pref_value)
        except ValueError:
            if preset_type == PresetPrefType.preset_photo_rename:
                t = "Photo"
            else:
                t = "Video"
            logging.error(
                "%s extension case value is invalid. Resetting to lower case.", t
            )
            index = 2
        self.extensionCombo.setCurrentIndex(index)
        self.extensionCombo.currentIndexChanged.connect(self.extensionChanged)

        self.example = QLabel()
        self.updateExampleFilename()

        layout.addRow(_("Preset:"), self.renameCombo)
        layout.addRow(_("Extension:"), self.extensionCombo)
        layout.addRow(_("Example:"), self.example)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

    def setRenameComboIndex(self) -> None:
        """
        Set the value being displayed in the combobox to reflect the
        current renaming preference.

        Takes into account built-in renaming presets and custom presets.
        """

        index = self.index_lookup(self.preset_pref_lists)
        if index == -1:
            # Set to the "Custom..." value
            cb_index = self.renameCombo.count() - 1
        else:
            # Set to the appropriate combobox index, allowing for possible separator
            cb_index = self.renameCombo.getComboBoxIndex(index)
        logging.debug(
            "Setting %s combobox chosen value to %s",
            self.file_type.name,
            self.renameCombo.itemText(cb_index),
        )
        self.renameCombo.setCurrentIndex(cb_index)

    def pref_list(self) -> list[str]:
        """
        :return: the user's file naming preference according to whether
         this widget is handling photos or videos
        """
        if self.preset_type == PresetPrefType.preset_photo_rename:
            return self.prefs.photo_rename
        else:
            return self.prefs.video_rename

    @pyqtSlot(int)
    def renameComboItemActivated(self, index: int) -> None:
        """
        Respond to user activating the Rename preset combo box.

        :param index: index of the item activated
        """

        user_pref_list = None

        preset_class = self.renameCombo.currentData()
        if preset_class == PresetClass.start_editor:
            prefDialog = PrefDialog(
                self.pref_defn,
                self.pref_list(),
                self.generation_type,
                self.prefs,
                self.sample_rpd_file,
            )

            if prefDialog.exec():
                user_pref_list = prefDialog.getPrefList()
                if not user_pref_list:
                    user_pref_list = None

            # Regardless of whether the user clicked OK or cancel, refresh the rename
            # combo box entries
            self.getCustomPresets()
            self.renameCombo.resetEntries(self.preset_names)
            self.setUserPrefList(user_pref_list=user_pref_list)
            self.setRenameComboIndex()
        else:
            assert (
                preset_class == PresetClass.custom
                or preset_class == PresetClass.builtin
            )
            index = self.renameCombo.getPresetIndex(self.renameCombo.currentIndex())
            user_pref_list = self.combined_pref_lists[index]
            self.setUserPrefList(user_pref_list=user_pref_list)

        self.updateExampleFilename()

    def getCustomPresets(self) -> None:
        """
        Get the custom presets from the user preferences and store them in lists
        """

        self.preset_names, self.preset_pref_lists = self.prefs.get_custom_presets(
            preset_type=self.preset_type
        )
        self.combined_pref_lists = self.pref_conv + tuple(self.preset_pref_lists)

    def setUserPrefList(self, user_pref_list: list[str]) -> None:
        """
        Update the user preferences with a new preference value
        :param user_pref_list: the photo or video rename preference list
        """

        if user_pref_list is not None:
            logging.debug("Setting new %s rename preference value", self.file_type.name)
            if self.preset_type == PresetPrefType.preset_photo_rename:
                self.prefs.photo_rename = user_pref_list
            else:
                self.prefs.video_rename = user_pref_list

    def updateExampleFilename(
        self,
        downloads_today: list[str] | None = None,
        stored_sequence_no: int | None = None,
    ) -> None:
        """
        Update filename shown to user that serves as an example of the
        renaming rule in practice on sample data.

        :param downloads_today: if specified, update the downloads today value
        :param stored_sequence_no: if specified, update the stored sequence value
        """

        if downloads_today:
            self.sample_rpd_file.sequences.downloads_today_tracker.downloads_today = (
                downloads_today
            )
        if stored_sequence_no is not None:
            self.sample_rpd_file.sequences.stored_sequence_no = stored_sequence_no

        if self.preset_type == PresetPrefType.preset_photo_rename:
            self.name_generator = gn.PhotoName(self.prefs.photo_rename)
            logging.debug("Updating example photo name in rename panel")
        else:
            self.name_generator = gn.VideoName(self.prefs.video_rename)
            logging.debug("Updating example video name in rename panel")

        self.example.setText(self.name_generator.generate_name(self.sample_rpd_file))

    def updateSampleFile(self, sample_rpd_file: Photo | Video) -> None:
        self.sample_rpd_file = make_sample_rpd_file(
            sample_rpd_file=sample_rpd_file,
            sample_job_code=self.prefs.most_recent_job_code(missing=_("Job Code")),
            prefs=self.prefs,
            generation_type=self.generation_type,
        )
        self.updateExampleFilename()

    @pyqtSlot(int)
    def extensionChanged(self, index: int) -> None:
        """
        Respond to user changing the case of file extensions in file name generation.

        Save new preference value, and update example file name.
        """

        value = self.extensionCombo.currentData()
        if self.preset_type == PresetPrefType.preset_photo_rename:
            self.prefs.photo_extension = value
        else:
            self.prefs.video_extension = value
        self.sample_rpd_file.generate_extension_case = value
        self.updateExampleFilename()


class RenameOptionsWidget(FlexiFrame):
    """
    Display and allow editing of preference values for Downloads today
    and Stored Sequence Number and associated options, as well as
    the strip incompatible characters option.
    """

    def __init__(
        self,
        prefs: Preferences,
        photoRenameWidget: RenameWidget,
        videoRenameWidget: RenameWidget,
        parent,
    ) -> None:
        super().__init__(parent=parent)

        self.prefs = prefs
        self.photoRenameWidget = photoRenameWidget
        self.videoRenameWidget = videoRenameWidget

        self.setBackgroundRole(QPalette.Base)
        self.setAutoFillBackground(True)

        compatibilityLayout = QVBoxLayout()
        layout = self.layout()

        # QSpinBox cannot display values greater than this value
        self.c_maxint = platform_c_maxint()

        tip = _("A counter for how many downloads occur on each day")
        self.downloadsTodayLabel = QLabel(_("Downloads today:"))
        self.downloadsToday = QSpinBox()
        self.downloadsToday.setMinimum(0)
        # QSpinBox defaults to a maximum of 99
        self.downloadsToday.setMaximum(self.c_maxint)
        self.downloadsToday.setToolTip(tip)

        # This instance of the downloads today tracker is secondary to the
        # instance in the rename files process. That process automatically
        # updates the value and then once a download is complete, the
        # downloads today value here is overwritten.
        self.downloads_today_tracker = DownloadsTodayTracker(
            day_start=self.prefs.day_start, downloads_today=self.prefs.downloads_today
        )

        downloads_today = self.downloads_today_tracker.get_or_reset_downloads_today()
        if self.prefs.downloads_today != self.downloads_today_tracker.downloads_today:
            self.prefs.downloads_today = self.downloads_today_tracker.downloads_today

        self.downloadsToday.setValue(downloads_today)
        self.downloadsToday.valueChanged.connect(self.downloadsTodayChanged)

        tip = _("A counter that is remembered each time the program is run ")
        self.storedNumberLabel = QLabel(_("Stored number:"))
        self.storedNumberLabel.setToolTip(tip)
        self.storedNumber = QSpinBox()
        self.storedNumberLabel.setBuddy(self.storedNumber)
        self.storedNumber.setMinimum(0)
        self.storedNumber.setMaximum(self.c_maxint)
        self.storedNumber.setToolTip(tip)

        self.storedNumber.setValue(self.stored_sequence_no)
        self.storedNumber.valueChanged.connect(self.storedNumberChanged)

        tip = _(
            "The time at which the <i>Downloads today</i> sequence number should be "
            "reset"
        )
        self.dayStartLabel = QLabel(_("Day start:"))
        self.dayStartLabel.setToolTip(tip)

        self.dayStart = QTimeEdit()
        self.dayStart.setToolTip(tip)
        self.dayStart.setTime(self.prefs.get_day_start_qtime())
        self.dayStart.timeChanged.connect(self.timeChanged)
        # 24 hour format, if wanted in a future release:
        # self.dayStart.setDisplayFormat('HH:mm:ss')

        self.sync = QCheckBox(_("Synchronize RAW + JPEG"))
        self.sync.setChecked(self.prefs.synchronize_raw_jpg)
        self.sync.stateChanged.connect(self.syncChanged)
        tip = _(
            "Synchronize sequence numbers for matching RAW and JPEG pairs.\n\n"
            "See the online documentation for more details."
        )
        self.sync.setToolTip(tip)

        self.sequences = QGroupBox(_("Sequence Numbers"))

        sequencesLayout = QFormLayout()

        sequencesLayout.addRow(self.storedNumberLabel, self.storedNumber)
        sequencesLayout.addRow(self.downloadsTodayLabel, self.downloadsToday)
        sequencesLayout.addRow(self.dayStartLabel, self.dayStart)
        sequencesLayout.addRow(self.sync)

        self.sequences.setLayout(sequencesLayout)

        self.stripCharacters = QCheckBox(_("Strip incompatible characters"))
        self.stripCharacters.setChecked(self.prefs.strip_characters)
        self.stripCharacters.stateChanged.connect(self.stripCharactersChanged)
        self.stripCharacters.setToolTip(
            _(
                "Whether photo, video and folder names should have any characters "
                "removed that are not allowed by other operating systems"
            )
        )
        self.compatibility = QGroupBox(_("Compatibility"))
        self.compatibility.setLayout(compatibilityLayout)
        compatibilityLayout.addWidget(self.stripCharacters)

        layout.addWidget(self.sequences)
        layout.addWidget(self.compatibility)
        layout.addStretch()
        layout.setSpacing(18)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)

    @property
    def stored_sequence_no(self) -> int:
        try:
            stored_value = int(self.prefs.stored_sequence_no) + 1
            assert 0 <= stored_value <= self.c_maxint
        except (ValueError, AssertionError):
            stored_value = 0
            logging.error("Resetting invalid stored sequence number to 0")
            self.prefs.stored_sequence_no = -1
        return stored_value

    @stored_sequence_no.setter
    def stored_sequence_no(self, value: int) -> None:
        logging.info("Setting stored sequence no to %d", value)
        self.prefs.stored_sequence_no = value - 1

    @pyqtSlot(QTime)
    def timeChanged(self, time: QTime) -> None:
        hour = time.hour()
        minute = time.minute()
        self.prefs.day_start = f"{hour}:{minute}"
        logging.debug("Setting day start to %s", self.prefs.day_start)
        self.downloads_today_tracker.set_day_start(hour=hour, minute=minute)

    @pyqtSlot(int)
    def downloadsTodayChanged(self, value: int) -> None:
        self.downloads_today_tracker.reset_downloads_today(value=value)
        dt = self.downloads_today_tracker.downloads_today
        logging.debug("Setting downloads today value to %s %s", dt[0], dt[1])
        self.prefs.downloads_today = dt
        if self.prefs.photo_rename_pref_uses_downloads_today():
            self.photoRenameWidget.updateExampleFilename(downloads_today=dt)
        if self.prefs.video_rename_pref_uses_downloads_today():
            self.videoRenameWidget.updateExampleFilename(downloads_today=dt)

    @pyqtSlot(int)
    def storedNumberChanged(self, value: int) -> None:
        self.stored_sequence_no = value
        if self.prefs.photo_rename_pref_uses_stored_sequence_no():
            self.photoRenameWidget.updateExampleFilename(stored_sequence_no=value - 1)
        if self.prefs.video_rename_pref_uses_stored_sequence_no():
            self.videoRenameWidget.updateExampleFilename(stored_sequence_no=value - 1)

    @pyqtSlot(int)
    def syncChanged(self, state: int) -> None:
        sync = state == Qt.Checked
        logging.debug("Setting synchronize RAW + JPEG sequence values to %s", sync)
        self.prefs.synchronize_raw_jpg = sync

    @pyqtSlot(int)
    def stripCharactersChanged(self, state: int) -> None:
        strip = state == Qt.Checked
        logging.debug("Setting strip incompatible characers to %s", strip)
        self.prefs.strip_characters = strip


class RenamePanel(ScrollAreaNoFrame):
    """
    Renaming preferences widget, for photos, videos, and general
    renaming options.
    """

    def __init__(self, parent) -> None:
        super().__init__(name="renamePanel", parent=parent)
        assert parent is not None
        self.rapidApp = parent
        self.prefs = self.rapidApp.prefs
        self.setObjectName("renamePanelScrollArea")

        self.photoRenamePanel = QPanelView(
            label=_("Photo Renaming"),
        )
        self.photoRenamePanel.setObjectName("photoRenamePanelView")

        self.videoRenamePanel = QPanelView(
            label=_("Video Renaming"),
        )
        self.videoRenamePanel.setObjectName("videoRenamePanelView")
        self.renameOptionsPanel = QPanelView(
            label=_("Renaming Options"),
        )
        self.renameOptionsPanel.setObjectName("renameOptionsPanelView")

        self.photoRenameWidget = RenameWidget(
            preset_type=PresetPrefType.preset_photo_rename,
            prefs=self.prefs,
            parent=self,
            exiftool_process=self.rapidApp.exiftool_process,
        )
        self.photoRenameWidget.setObjectName("photoRenameWidget")

        self.videoRenameWidget = RenameWidget(
            preset_type=PresetPrefType.preset_video_rename,
            prefs=self.prefs,
            parent=self,
            exiftool_process=self.rapidApp.exiftool_process,
        )
        self.videoRenameWidget.setObjectName("videoRenameWidget")

        self.renameOptions = RenameOptionsWidget(
            prefs=self.prefs,
            parent=self,
            photoRenameWidget=self.photoRenameWidget,
            videoRenameWidget=self.videoRenameWidget,
        )
        self.renameOptions.setObjectName("renameOptionsWidget")

        self.photoRenamePanel.addWidget(self.photoRenameWidget)
        self.videoRenamePanel.addWidget(self.videoRenameWidget)
        self.renameOptionsPanel.addWidget(self.renameOptions)

        for widget in (
            self.photoRenameWidget,
            self.videoRenameWidget,
            self.renameOptions,
        ):
            self.verticalScrollBarVisible.connect(widget.containerVerticalScrollBar)

        self.horizontalScrollBarVisible.connect(
            self.renameOptions.containerHorizontalScrollBar
        )

        widget = QWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(QSplitter().handleWidth())
        widget.setLayout(layout)
        layout.addWidget(self.photoRenamePanel)
        layout.addWidget(self.videoRenamePanel)
        layout.addWidget(self.renameOptionsPanel)

        self.setWidget(widget)
        self.setWidgetResizable(True)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

    def updateSequences(
        self, downloads_today: list[str], stored_sequence_no: int
    ) -> None:
        """
        Update the value displayed in the display to reflect any values changed after
        the completion of a download.

        :param downloads_today: new downloads today value
        :param stored_sequence_no: new stored sequence number value
        """

        self.renameOptions.downloadsToday.setValue(int(downloads_today[1]))
        self.renameOptions.downloads_today_tracker.downloads_today = downloads_today
        self.renameOptions.storedNumber.setValue(stored_sequence_no + 1)

    def setSamplePhoto(self, sample_photo: Photo) -> None:
        self.photoRenameWidget.updateSampleFile(sample_rpd_file=sample_photo)

    def setSampleVideo(self, sample_video: Video) -> None:
        self.videoRenameWidget.updateSampleFile(sample_rpd_file=sample_video)
